import pathlib

import pytest

from pygrambank.cldf import GlottologGB
from pygrambank import sheet


@pytest.fixture
def sheet_abbr(api):
    return sheet.Sheet(api.sheets_dir / 'ABBR_abcd1234.tsv')


@pytest.fixture
def sheet_factory(tmp_path):
    def _make_one(text, name='ABBR_bcde1234.tsv'):
        fname = tmp_path / name
        fname.write_text(text, encoding='utf8')
        return sheet.Sheet(fname)
    return _make_one


def test_visitor(sheet_abbr):
    class Counter(object):
        def __init__(self):
            self.c = 0

        def __call__(self, *args, **kw):
            self.c += 1

    c = Counter()
    sheet_abbr.visit(c)
    assert c.c == 2


def test_check(api, sheet_abbr, sheet_factory, capsys):
    assert sheet_abbr.check(api) >\
           sheet.Sheet(api.sheets_dir / 'NOVALS_abcd1234.tsv').check(api)


@pytest.mark.filterwarnings("ignore:Duplicate")
@pytest.mark.parametrize(
    'text,error',
    [
        ('Value', 'missing column'),
        ('Value\tValue', 'duplicate header'),
        ('Feature_ID\tValue\tSource\tComment\t\nGB020\t1\t\t\tx', 'non-empty cell'),
        ('Feature_ID\tValue\tSource\tComment\t\nGB020\t1\t\t\tx\nGB020\t0\t\t\t', 'inconsistent'),
    ]
)
def test_check2(api, sheet_factory, text, error, capsys):
    s = sheet_factory(text)
    s.check(api)
    out, _ = capsys.readouterr()
    assert error in out


def test_values(sheet_abbr, api):
    assert len(list(sheet_abbr.iter_row_objects(api))) == 1


def test_Sheet(api, capsys, sheet_factory):
    s = sheet_factory("""Feature_ID\tValue\tSource\tComment\tContributed datapoints
GB020\t0\tMeier 2007\t\tHJH\n""")
    assert str(s).endswith('tsv')
    rows = list(s.iter_row_objects(api))
    assert rows[0].sources[0].key == ('Meier', '2007', '')
    assert len(rows) == 1
    md = s.metadata(GlottologGB(pathlib.Path(__file__).parent / 'glottolog'))
    out, _ = capsys.readouterr()
    assert 'no macroareas' in out
    assert md['Language_level_ID'] == 'abcd1234'
    assert s.visit() == (1, 1)


@pytest.mark.parametrize(
    'row,logs',
    [
        (dict(Feature_ID='XX056', Value='1', Source='Ref'), True),
        (dict(Feature_ID='GB020', Value='x', Source='Ref', Comment=''), True),
        (dict(Feature_ID='GB020', Value='', Source='Ref', Comment=''), True),
        (dict(Feature_ID='GB020', Value='', Source='', Comment='text'), True),
        (dict(Feature_ID='', Value='1', Source='Ref'), False),
    ]
)
def test_valid_row(sheet_abbr, api, mocker, row, logs):
    log = mocker.Mock()
    assert not sheet_abbr.valid_row(row, api, log=log)
    assert (not logs) or log.called



"""
Check for non-UTF-8 encoding.

Also checks for Unicode Replacement Characters, because those tend to be a sign
that data corruption might have happened.
"""

import sys

from clldutils.clilib import PathType


REPLACEMENT_CHAR = chr(0xfffd)
SUGGESTED_ENCS = ['cp1252', 'macroman']


def register(parser):
    parser.add_argument('path', nargs='+', type=PathType(type='file'))


def _find_consecutive(pred, iterable):
    current_start = None
    for index, elem in enumerate(iterable):
        if pred(elem):
            if current_start is None:
                current_start = index
        else:
            if current_start is not None:
                yield current_start, index
                current_start = None
    if current_start:
        yield current_start, index
        current_start = None


def context(line, start, end, context_len):
    """Return substring from `start` to `end` (excl.) plus `context_len`
    characters on each side.

    Also works with raw `bytes`.
    """
    return line[max(0, start - context_len):min(len(line), end + context_len)]


def find_replacement_chars(line):
    """Find instances of Unicode Replacement Characters in `line`."""
    for start, end in _find_consecutive(lambda c: c == REPLACEMENT_CHAR, line):
        yield context(line, start, end, 20)


def find_non_ascii(raw_line):
    """Find characters outside of the ASCII range."""
    for start, end in _find_consecutive(lambda b: b >= 0x80, raw_line):
        yield context(raw_line, start, end, 20)


def suggest_encodings(raw_line):
    """Try out a few character encodings and see what the result looks like.

    Returns dictionary (encoding -> result).
    """
    suggestions = {}
    for enc in SUGGESTED_ENCS:
        try:
            suggestions[enc] = raw_line.decode(enc)
        except UnicodeError:  # pragma: nocover
            suggestions[enc] = '<could not decode from {}>'.format(enc)
    return suggestions


def run(args):
    for p in args.path:
        with open(p, 'rb') as f:
            raw_lines = list(f)

        utf8_detected = False
        nonutf8_detected = False

        for index, raw_line in enumerate(raw_lines):
            raw_line = raw_line.rstrip(b'\r\n')
            lineno = index + 1
            try:
                unicode_line = raw_line.decode('utf-8')
                # check for non-ascii characters
                if any(b >= 0x80 for b in raw_line):
                    utf8_detected = True
                for msg in find_replacement_chars(unicode_line):
                    print(
                        '{}:{}: Unicode replacement character found: {}'.format(
                            p, lineno, msg),
                        file=sys.stderr)
            except UnicodeDecodeError:
                nonutf8_detected = True
                for match in find_non_ascii(raw_line):
                    print('{}:{}: Non-UTF8 character found: {}'.format(
                        p, lineno, repr(match)))
                    for enc, res in sorted(suggest_encodings(match).items()):
                        print(" * {}:\t'{}'".format(enc, res.rstrip('\r\n')))

        if utf8_detected and nonutf8_detected:
            print(
                '{}: Mixed encoding detected! Tread with caution!'.format(p),
                file=sys.stderr)




def test_Issue(api):
    issue = api.issues[0]
    assert issue.html.startswith('<p>')
    assert list(issue.iter_urls())
    assert list(issue.iter_issue_links())
    assert list(issue.iter_links())[0] == ('http://example.com', 'ex')
    assert list(issue.iter_images())[0] == ('http://example.org/image.jpg', 'alt')



import re
import collections

import attr
from clldutils.markup import iter_markdown_tables

PHOTO_URI = 'https://glottobank.org/photos/{Photo}'
ROLES = [
    'Project leader',
    'Project coordinator',
    'Database manager',
    'Patron',
    'Node leader',
    'Coder',
    'Methods-team',
    'Senior advisor',
]


def parse_photo(s):
    match = re.search(r'src="(?P<url>[^"]+)"', s)
    if match:
        return match.group('url').split('/')[-1]


def parse_roles(s):
    return [r.strip() for r in s.split(',') if r.strip()]


def valid_roles(instance, attribute, value):
    validator = attr.validators.in_(ROLES)
    for vv in value:
        validator(instance, attribute, vv)


@attr.s
class Contributor(object):
    id = attr.ib()
    last_name = attr.ib()
    first_name = attr.ib()
    contribution = attr.ib(converter=parse_roles, validator=valid_roles)
    node = attr.ib(default='')
    status = attr.ib(default='')
    language_competence = attr.ib(default='')
    github_username = attr.ib(default='')
    email = attr.ib(default='')
    photo = attr.ib(converter=parse_photo, default='')
    bio = attr.ib(default='')
    should_be_included_despite_no_coding = attr.ib(default=False)

    @property
    def name(self):
        return '{0.first_name} {0.last_name}'.format(self)

    @property
    def roles(self):
        return sorted(self.contribution, key=lambda r: ROLES.index(r))


def norm_header(s):
    return s.lower().replace(' ', '_').replace('-', '_')


class Contributors(list):
    @classmethod
    def from_md(cls, fname):
        header, rows = next(iter_markdown_tables(fname.read_text(encoding='utf8')))
        header = [norm_header(c) for c in header]
        rows = [
            Contributor(**dict(zip(header, row)))
            for row in rows]
        byid = collections.Counter(r.id for r in rows)
        if byid.most_common(1)[0][1] > 1:  # pragma: no cover
            raise ValueError(
                'duplicate ids: {0}'.format([k for k, v in byid.most_common() if v > 1]))
        return cls(rows)
